/**
 * Title:        Grid Broker
 * Version:      1.0, 12/09/2001
 * @version      1.0, Sep 12, 2001, full version released: Dec. 2001
 * Description:  A Grid Scheduler for Application Scheduling on Grid based on
 *               Deadline and Budget Constrained Scheduling Algorithms
 * Authors:      Rajkumar Buyya and Manzur Murshed
 *               Monash University, Melbourne, Australia
 * Licence:      GPL - http://www.gnu.org/copyleft/gpl.html
 */

import java.util.*;
import java.io.*;
import eduni.simjava.*;
import gridsim.*;

/**
 * A class that maintains Resource details and Gridlets assigned to it.
 * It is used extensively in GridBroker class.
 */
class BrokerResource
{
  public ResourceCharacteristics resource;
  public GridletList glList;
  public double LatestLoad;
  public double AvailableMIPS;
  public int NoOfGridletsFinishedSoFar;
  public int NoOfGridletsDispatchedSoFar;

  public BrokerResource(ResourceCharacteristics resource, double LatestLoad)
  {
    this.resource = resource;
    glList = new GridletList();
    this.LatestLoad = LatestLoad;
    AvailableMIPS = resource.GetMIPSRating();
    NoOfGridletsFinishedSoFar = 0;
    NoOfGridletsDispatchedSoFar = 0;
  }

  /**
   * It returns Number of System Time Units (say Sec.) from
   * current time to the EndTime (say Deadline)
   */

  public double RemainingTimePeriodFromNow( double EndTime )
  {
    return EndTime - Sim_system.clock();
  }

  /**
   * This module checks available MI on a single PE within available deadline period
   * is sufficient for a Gridlet to run successful. if yes, it returns TRUE otherwise, false.
   */
  public boolean IsOKtoAssignGridletToResource( Gridlet gl, double EndTime )
  {
    double GridletNeedMI = gl.GetGridletLength();
    double TimeSpan =	RemainingTimePeriodFromNow( EndTime );
    double AvailableMIonOnePE = resource.GetMIPSRatingOfOnePE() * TimeSpan;

    if( AvailableMIonOnePE >= GridletNeedMI )
      return true;
    else
      return false;
  }

  /**
   * Returns available MIs depending on resource load from current time.
   */
  public double GetAvailableMI(double EndTime)
  {
    double TimeSpan = RemainingTimePeriodFromNow( EndTime );
    double AvailableMI = AvailableMIPS * TimeSpan;
    for(int i = 0; i < glList.size(); i++)
    {
      Gridlet gl = (Gridlet) glList.get(i);
      int status = gl.GetStatus();
      if(status!=Gridlet.READY)
        AvailableMI -= gl.GetGridletLength();
    }
    return AvailableMI;
  }

  public int NoOfGridletsInReady()
  {
    int count = 0;
    for(int i = 0; i<glList.size(); i++)
      if( ((Gridlet) glList.get(i)).GetStatus() ==  Gridlet.READY)
        count++;
    return count;
  }

  public int NoOfGridletsInQueue() {
    int count = 0;
    for(int i = 0; i<glList.size(); i++) {
      if( ((Gridlet) glList.get(i)).GetStatus() ==  Gridlet.QUEUED)
        count++;
    }
    return count;
  }
  public int NoOfGridletsInExec() {
    int count = 0;
    for(int i = 0; i<glList.size(); i++) {
      if( ((Gridlet) glList.get(i)).GetStatus() ==  Gridlet.INEXEC)
        count++;
    }
    return count;
  }

  public int NoOfGridletsFinished() {
    int count = 0;
    for(int i = 0; i<glList.size(); i++) {
      int status = ((Gridlet) glList.get(i)).GetStatus();
      if(status==Gridlet.SUCCESS || status==Gridlet.FAILED)
        count++;
    }
    return count;
  }
  /**
   * Identify an expected processing cost on the resource
   * @param gl Gridlet object.
   * @return an expected processing cost on the resource
   */
  public double ExpectedProcessingCost(Gridlet gl)
  {
    return gl.GetGridletLength()*resource.GetCostPerMI();
  }

}

/** The Grid Resource Broker: It is a GridSim entity with network.
 *  It takes user experiment and assigns Gridlets/Jobs to Grid resources
 * depending on scheduling strategy and deadline and budget constrains.
 */
class Broker extends GridSim
{
  private Experiment experiment;   // contains application, user requirements etc.
  private LinkedList ResIDList;    // Resource Enity ID List
  private LinkedList BrokerResourceList;
  private GridletList glUnfinishedList;
  private GridletList glFinishedList;
  private int GridletDispatched;
  private int GridletReturned;
  private double Expenses;
  private int MaxGridletPerPE;

  public Broker(String name, double baud_rate)
  {
    super(name, baud_rate);
    GridletDispatched = 0;
    GridletReturned = 0;
    Expenses = 0.0;
    MaxGridletPerPE = 2;
  }

  class OrderCost implements Comparator
  {
    public OrderCost() {
      super();
    }
    public int compare(Object a, Object b) {
      BrokerResource BRa = (BrokerResource) a;
      BrokerResource BRb = (BrokerResource) b;

      Double d_c1 = new Double(BRa.resource.GetCostPerMI());
      Double d_c2 = new Double(BRb.resource.GetCostPerMI());

      return d_c1.compareTo(d_c2);
    }
  }

  class OrderCostTime implements Comparator
  {
    public OrderCostTime() {
      super();
    }
    public int compare(Object a, Object b) {
      BrokerResource BRa = (BrokerResource) a;
      BrokerResource BRb = (BrokerResource) b;

      Double d_c1 = new Double(BRa.resource.GetCostPerMI());
      Double d_c2 = new Double(BRb.resource.GetCostPerMI());
      Double d_t1 = new Double(BRa.resource.GetMIPSRating()*(1-BRa.LatestLoad));
      Double d_t2 = new Double(BRb.resource.GetMIPSRating()*(1-BRb.LatestLoad));

      if(d_c1.equals(d_c2))
          return d_t2.compareTo(d_t1);
      else
        return d_c1.compareTo(d_c2);
    }
  }

  public void body()
  {
    Sim_event ev = new Sim_event();

    // Accept User Commands and Process
    for( sim_get_next(ev); ev.get_tag()!=GridSimTags.END_OF_SIMULATION; sim_get_next(ev))
    {
      // System.out.println(get_name()+"clock: "+Sim_system.clock()+" : Queue Size-B (after reading Expt event): " + "user: "+sim_waiting(new Sim_from_port(user))+" input: "+sim_waiting(new Sim_from_port(input)));
      // Extra Experiment Detatils from the User Event
      experiment = (Experiment) ev.get_data();
      int UserEntityID = ev.get_src();

      // Set GridLet's Owner ID as User Broker ID
      for(int i=0; i<experiment.GetGridletList().size(); i++)
        ((Gridlet) experiment.GetGridletList().get(i)).SetUserID(get_id());

      /* RESOURCE DISCOVERY: Get Resource List from GIS + Property from Resource Info Service */

      // Requesting the GridInformationService to send a list of resources
      ResIDList = GetGridResourceList();
      RecordStatistics("BROKER.Resource.Count", ResIDList.size());

      // Cretae Resource List with Static and Dynamic Info.
      BrokerResourceList = new LinkedList();
      Accumulator accMIPSRatingAll = new Accumulator();
      Accumulator accMIPSRating = new Accumulator();
      Accumulator accMIPSRatingAllWithLoad = new Accumulator();
      Accumulator accCostPerMI = new Accumulator();

      // Get Resource Static and Dynamic Characteristics
      for( int i=0; i < ResIDList.size(); i++)
      {
        // Get Resource ID
        int res_id = ((Integer) ResIDList.get(i)).intValue();

        // Get Resource Characteristic Info: Send Request and Receive Event/Msg
        Send(output, GridSimTags.SCHEDULE_NOW, GridSimTags.RESOURCE_CHARACTERISTICS, new IO_data(new Integer(get_id()), 4, res_id));
        ResourceCharacteristics res = (ResourceCharacteristics) ReceiveEventObject(input);

        accMIPSRating.add(res.GetMIPSRatingOfOnePE());
        accMIPSRatingAll.add(res.GetMIPSRating());
        accCostPerMI.add(res.GetCostPerMI());

        // Get Resource Dynamic information
        Send(output, GridSimTags.SCHEDULE_NOW, GridSimTags.RESOURCE_DYNAMICS, new IO_data(new Integer(get_id()), 4, res_id));
        Accumulator accLoad = (Accumulator) ReceiveEventObject(input);

        accMIPSRatingAllWithLoad.add(res.GetMIPSRating()*(1-accLoad.mean()));
        BrokerResourceList.add(new BrokerResource(res, accLoad.mean()));
      }

      Accumulator accLength = experiment.GetAccLength();
      double total_MI = accLength.sum();

      double MAX_time = total_MI/accMIPSRatingAllWithLoad.min();
      double MIN_time = Math.max(total_MI/accMIPSRatingAll.sum(),
                        accLength.max()/accMIPSRating.max());
      double deadline = experiment.GetD_factor()*(MAX_time-MIN_time) + MIN_time;

      double MAX_cost = total_MI*accCostPerMI.max();
      double MIN_cost = total_MI*accCostPerMI.min();
      double budget = experiment.GetB_factor()*(MAX_cost-MIN_cost) + MIN_cost;

      System.out.println(get_name()+":: Deadline: "+deadline+" Budget: "+budget);
      experiment.SetBudget(budget);
      experiment.SetDeadline(deadline + Sim_system.clock());

      RecordStatistics("USER.Experiment.Budget", budget);
      RecordStatistics("USER.Experiment.Deadline", deadline);

      glUnfinishedList = (GridletList) experiment.GetGridletList().clone();
      glFinishedList = new GridletList();

      GridletDispatched = 0;
      GridletReturned = 0;

      Expenses = 0.0;

      // SCHEDUDLING: Do until experiment finishes
      while (glFinishedList.size() < experiment.GetGridletList().size())
      {
        if( (GridSim.Clock() >= experiment.GetDeadline()) || (Expenses >= experiment.GetBudget()) )
          break;

        int scheduled = ScheduleAdviser();
        int dispatched = Dispatcher();
        int received = ReceiveFinishedGridlets();

        // heurisitics for deciding hold condition
        if(dispatched<=0 && received<=0 && glUnfinishedList.size()>0)
        {
          double deadline_left = experiment.GetDeadline()-GridSim.Clock();
          GridSimHold(Math.max(deadline_left*0.01, 1.0));
        }
      }

      // Capture all events related to Gridlets that have been scheduled
      // for execution, but not yet received.
      // NOTE: Do not use (GridletDispatched-GridletReturned) in a loop
      // since they are Global Variables in GridBroker entity.
      // ERROR LINE:  for(int i = 0; i < GridletDispatched-GridletReturned; i++)
      while( GridletDispatched > GridletReturned )
        ReceiveFinishedGridlets();

      // inform the User Entity that Broker completed processing
      Send(UserEntityID, GridSimTags.SCHEDULE_NOW, GridSimTags.EXPERIMENT, experiment);
    }
//    System.out.println( get_name()+" is termindated");
    TerminateInputOutputEntities();
  }

  private int ReceiveFinishedGridlets()
  {
    int received = 0;

    if(GridletDispatched > GridletReturned)
    {
      do
      {
        // Get Gridlet finished information
        Gridlet glFinished = (Gridlet) ReceiveEventObject(input);

        RecordStatistics("BROKER.Gridlet.EndTime", glFinished.GetGridletID());

  //System.out.println(glFinished.GetGridletID() + " Finish==> " + (int)Sim_system.clock());
        BrokerResource brServed = WhoServed(glFinished);
        brServed.NoOfGridletsFinishedSoFar++;

        double MIPSPerGridlet = glFinished.GetGridletLength()/glFinished.GetWallClockTime();
        brServed.AvailableMIPS = MIPSPerGridlet*MaxGridletPerPE*brServed.resource.GetNoOfPEs();
  //System.out.println(Sim_system.get_entity(brServed.resource.GetID()).get_name() + ": " + brServed.AvailableMIPS + " " + brServed.NoOfGridletsInExec() + " " + brServed.NoOfGridletsFinished());
        brServed.glList.remove(glFinished);
        glFinishedList.add(glFinished);
        GridletReturned++;
        Expenses += glFinished.GetActualCPUTime()*glFinished.GetCostPerSec();
        received++;
      } while(sim_waiting(new Sim_from_port(input))>0);
    }
    return received;
  }

  // return Broker Resource that executed Gridlet
  private BrokerResource WhoServed(Gridlet gl)
  {
    BrokerResource br = null;
    for(int i = 0; i < BrokerResourceList.size(); i++)
    {
      br = (BrokerResource) BrokerResourceList.get(i);
      if(br.glList.contains(gl))
        break;
    }
    return br;
  }

  /**
   * Processing Cost of Gridlets in READY, INQUEUE, or INEXEC
   * @return Expected cost of processing Gridlets that are already committeed to resources
   */
  private double BudgetOnHold()
  {
    double amount = 0;
    for(int i=0; i < BrokerResourceList.size(); i++)
    {
      BrokerResource br = (BrokerResource) BrokerResourceList.get(i);
      for(int j=0; j<br.glList.size(); j++ )
      {
        Gridlet gl = (Gridlet) br.glList.get(j);
        int status =  gl.GetStatus();
        if( status == Gridlet.READY || status == Gridlet.QUEUED || status == Gridlet.INEXEC )
            amount += br.ExpectedProcessingCost(gl);
      }
    }
    return amount;
  }

  /**
   * Decide whether budget is available for processing Gridlet on a broker resource
   * @param Gridlet
   * @param Broker Resource
   * @return true if budget available, false, otherwise.
   */
  private boolean IsBudgetAvailable(Gridlet gl, BrokerResource br)
  {
    if( (Expenses+BudgetOnHold()+br.ExpectedProcessingCost(gl)) <= experiment.GetBudget())
      return true;
    else
      return false;
  }

  /**
   * Depending on Optimisation Policy, it selects resources.
   * @return It returns count of number of Gridlets scheduled.
   */

  /*--------------------------------------------------------------------------*/
  private int ScheduleAdviser()
  {
    // Removing all non-dispatched gridlets
    // May be we can do Prediction to see how many to REMOVE from the previous assignment
    for(int i=0; i<BrokerResourceList.size(); i++)
    {
      BrokerResource br = (BrokerResource) BrokerResourceList.get(i);
      for(int j = 0; j < br.glList.size(); j++)
      {
        Gridlet gl = (Gridlet) br.glList.get(j);
        if(gl.GetStatus()==Gridlet.READY)
        {
          br.glList.remove(gl);
          glUnfinishedList.add(gl);
        }
      }
    }

    // Isn't below formulae gives ZERO ?
    int scheduled = experiment.GetNoOfGridlets()-glFinishedList.size()-glUnfinishedList.size();

    if(glUnfinishedList.size() == 0)
      return 0; // Nothing to schedule OR should it return "scheduled"

    // sort by the ascending order of length to maximize packing AND it can also be reverse to
    // minimise the risk of exceeding too much beyond deadline if resources do not perform as expected!
    glUnfinishedList.sort();

    switch(experiment.GetOptimizationStrategy())
    {
      case Experiment.OPTIMIZE_COST:
//				System.out.println("DBC OPTIMIZE_COST Scheduling Selected");
        scheduled += ScheduleWithDBC_CostOptimisation( new OrderCost());
        break;

      case Experiment.OPTIMIZE_COST_PLUS:
//				System.out.println("DBC OPTIMIZE_COST_PLUS Scheduling Selected");
        scheduled += ScheduleWithDBC_CostOptimisation( new OrderCostTime());
        break;

      case Experiment.OPTIMIZE_COST_TIME:
//				System.out.println("OPTIMIZE_COST_TIME Scheduling Selected");
        scheduled += ScheduleWithDBC_CostTimeOptimisation();
        break;

      case Experiment.OPTIMIZE_TIME:
//				System.out.println("OPTIMIZE_TIME Scheduling Selected");
        break;

      case Experiment.OPTIMIZE_NONE:
//				System.out.println("OPTIMIZE_NONE Scheduling Selected");
        break;
    }
    return scheduled;
  }

  /**
   *  It performs Deadline and Budget Constratined (DBC) Cost Minimisation Scheduling.
   *  @return It returns count of number of Gridlets scheduled
   */
  private int ScheduleWithDBC_CostOptimisation(Comparator sortComparator)
  {
    int scheduled = 0; // No of Gridlets scheduled during this iteration
    Collections.sort(BrokerResourceList, sortComparator);  // IS THIS REQUIRED: JUST RAJ ADDED
    for(int i=0; glUnfinishedList.size() > 0 && i<BrokerResourceList.size(); i++)
    {
      BrokerResource br = (BrokerResource) BrokerResourceList.get(i);
      double AvailableMI = br.GetAvailableMI(experiment.GetDeadline());
      GridletList glTempList = (GridletList) glUnfinishedList.clone();
      // schedule as long as there Gridlets and MIs are available!
      for(int j=0; j<glTempList.size() && AvailableMI > 0; j++)
      {
        Gridlet gl = (Gridlet) glTempList.get(j);
        // assign only when budget is available for processing a resource
        if(IsBudgetAvailable(gl, br))
        {
          if( br.IsOKtoAssignGridletToResource( gl, experiment.GetDeadline()) )
          {
            gl.SetStatus(Gridlet.READY);
            gl.SetResParam(br.resource.GetID(), br.resource.GetCostPerSec());
            glUnfinishedList.remove(gl);
            br.glList.add(gl);
            AvailableMI -= gl.GetGridletLength();
            scheduled++;
          }
        }
      }
    }
    return scheduled;
  }

  /**
   * It performs Deadline and Budget Constratined (DBC) Cost+Time Minimisation Scheduling.
   * @return It returns count of number of Gridlets scheduled
   */
  private int ScheduleWithDBC_CostTimeOptimisation()
  {
    int scheduled = 0; // No of Gridlets scheduled during this iteration
    Collections.sort(BrokerResourceList, new OrderCostTime());

    // Initialise BrokerResourcesWithSameCostPerMI parameters
    int currentBRnum = 0;
    BrokerResource br = (BrokerResource) BrokerResourceList.get(currentBRnum);
    LinkedList BrokerResourcesWithSameCostPerMI = new LinkedList();
    double AvailableMI = 0.0;

    while(glUnfinishedList.size()>0 && currentBRnum < BrokerResourceList.size())
    {
      // Create a List of BrokerResourcesWithSameCostPerMI
      br = (BrokerResource) BrokerResourceList.get(currentBRnum);
      BrokerResourcesWithSameCostPerMI.clear();
      // add current resource to the list
      BrokerResourcesWithSameCostPerMI.add((BrokerResource) br);
      currentBRnum++;
      AvailableMI = br.GetAvailableMI(experiment.GetDeadline());;

      // group all those resources with the same cost as current resource by adding to list BrokerResourcesWithSameCostPerMI
      while(currentBRnum < BrokerResourceList.size()-1 &&
        ((BrokerResource) BrokerResourceList.get(currentBRnum+1)).resource.GetCostPerMI() == br.resource.GetCostPerMI())
      {
        currentBRnum++;
        br = (BrokerResource) BrokerResourceList.get(currentBRnum);

        AvailableMI += br.GetAvailableMI(experiment.GetDeadline());
        BrokerResourcesWithSameCostPerMI.add((BrokerResource) br);
      }

      // Allocate Tasks to BrokerResourcesWithSameCostPerMI with Time Minimisation Style
      // schedule as long as there Gridlets and MIs are available!
      // THIS MODULE NEED MORE WORK: RAJ.....
      GridletList glTempList = (GridletList) glUnfinishedList.clone();
      for(int j=0; j<glTempList.size() && AvailableMI > 0 ; j++)
      {
        Gridlet gl = (Gridlet) glTempList.get(j);
        // assign only when budget is available for processing a resource
        if(IsBudgetAvailable(gl, br))
        {
          if( br.IsOKtoAssignGridletToResource( gl, experiment.GetDeadline()) )
          {
            gl.SetStatus(Gridlet.READY);
            gl.SetResParam(br.resource.GetID(), br.resource.GetCostPerSec());
            glUnfinishedList.remove(gl);

            int brIndex = 0;
            int glIndex = j;
            do
            {
              br = (BrokerResource) BrokerResourcesWithSameCostPerMI.get(brIndex);
              glIndex -= br.resource.GetNoOfPEs();
              if(++brIndex >= BrokerResourcesWithSameCostPerMI.size())
                brIndex = 0;
            } while(glIndex>=0);

            br.glList.add(gl);
            AvailableMI -= gl.GetGridletLength();
            scheduled++;
          }
        }
      }
    }
    return scheduled;
  }

  /**
   * The Dispatcher() is responsible for assigning Gridlets from
   * Gridlets list in Broker Resource to Grid Resource in such a way
   * that GridResource is not overload. This decision is made based
   * current load on Grid resource and size along in case of Time shared resources.
   * If resource is Space shared, then it is based on number of nodes that are free.
   */
  private int Dispatcher()
  {
    int dispatched = 0;
    //System.out.println("-------------------------------------");
    for(int i=0; i < BrokerResourceList.size(); i++)
    {
      BrokerResource br = (BrokerResource) BrokerResourceList.get(i);
      int NoOfPEs = br.resource.GetNoOfPEs();
      int NoOfGridletsOnResource = br.NoOfGridletsInExec()+ br.NoOfGridletsInQueue();;
      int NoOfGridletsToBeDispatched = br.NoOfGridletsInReady();

      int OptimalDisptachSize = 0;
      // Determine How many Gridlets to be dispatched depending on Type of Resource
      switch(br.resource.GetAllocationPolicy())
      {
        case ResourceCharacteristics.TIME_SHARED:
          // How many to be sent to a resource can be based on the following:
          // 1. Load on PEs or resource
          // 2. How many of your Gridlets are running
          // A simple heuristic can be be about 2 Gridlets on each PE
          // GET IDEA from Nimrod-G or some Heuristic in Laterature
          int NoOfGridletsCanBeDispatched = NoOfPEs*MaxGridletPerPE-NoOfGridletsOnResource;
          OptimalDisptachSize = Math.min(NoOfGridletsToBeDispatched, NoOfGridletsCanBeDispatched);
          break;

        case ResourceCharacteristics.SPACE_SHARED:
          // Need to look into many parameters: GET IDEA from Nimrod-G
          break;
        case ResourceCharacteristics.ADVANCE_RESERVATION:
          // Simply Dispatch and Bind to the Resource
          break;
      }
// System.out.println(OptimalDisptachSize);
      for(int j=0; j<br.glList.size() && OptimalDisptachSize>0; j++)
      {
        Gridlet gl = (Gridlet) br.glList.get(j);
        if(gl.GetStatus()==Gridlet.READY)
        {
          gl.SetStatus(Gridlet.QUEUED);
// System.out.println(gl.GetGridletID() + " Start with [" + Sim_system.get_entity(br.resource.GetID()).get_name() + "] ==> " + (int)Sim_system.clock());
          Send(output, GridSimTags.SCHEDULE_NOW, GridSimTags.GRIDLET_SUBMIT,
                      new IO_data(gl, gl.GetGridletFileSize(), br.resource.GetID()));

// System.out.println((int) Sim_system.clock() + " " + gl.GetGridletID() + ":" + gl.GetResourceID());
          br.NoOfGridletsDispatchedSoFar++;
          GridletDispatched++;
          OptimalDisptachSize--;
          dispatched++;
        }
      }
    }
    return dispatched;
  }
}
