/**
 * Title:        GridSim Toolkit
 * Version:      GridSim  1.0, 12/09/2001
 * Description:  GridSim (Grid Simulation) Toolkit for Modeling and Simulation of
 *               Parallel and Distributed Systems such as Clusters and Grids
 * @author       R. Buyya and M. Murshed (by surname)
 *               Monash University, Melbourne, Australia
 * Licence:      GPL - http://www.gnu.org/copyleft/gpl.html
 */

package gridsim;
import java.util.*;
import java.io.*;
import eduni.simjava.*;

/**
 * Gridlets at Grid Resource.
 * It maintains Gridlets along with details arrival and departure time with
 * machine and PE on which they are executed.
 */
class ResGridlet
{
  private Gridlet gridlet;
  private double arrival_time;
  private double finished_time;
  private double gridlet_finished_so_far;
  private int machine_id;
  private int PE_id;

  public ResGridlet(Gridlet gridlet)
  {
    this.gridlet = gridlet;
    this.arrival_time = Sim_system.clock();
    this.finished_time = -1;
    this.gridlet_finished_so_far = 0.0;
    this.machine_id = -1;
    this.PE_id = -1;
  }

  public void SetIDs(int machine_id, int PE_id)
  {
    this.machine_id = machine_id;
    this.PE_id = PE_id;
  }

  public double GetRemainingLength()
  {
    return gridlet.GetGridletLength() - gridlet_finished_so_far;
  }

  public void UpdateGridletFinishedSoFar(double mi_length)
  {
    gridlet_finished_so_far += mi_length;
  }

  public double GetArrivalTime()
  {
    return arrival_time;
  }

  public void SetFinishTime(double MIRate, ResourceCalendar rCalendar)
  {
    double fTime = GetRemainingLength()/MIRate + Sim_system.clock();
    Calendar finish = rCalendar.CalendarAtGivenSimTime(fTime);
    Calendar NextHour = rCalendar.CalendarNow();
    NextHour.add(Calendar.HOUR, 1);
    NextHour.set(Calendar.MINUTE, 0);
    NextHour.set(Calendar.SECOND, 0);

    if(finish.before(NextHour))
      finished_time = fTime;
    else
      finished_time = -1;
  }

  public double GetFinishTime()
  {
    return finished_time;
  }

  public Gridlet GetGridlet()
  {
    return gridlet;
  }
}

/**
 * A ResourceGridlet list. An extended LinkedList with
 * its own move function.
 */

 class ResGridletList extends LinkedList
{
  public ResGridletList()
  {
    super();
  }

  public boolean move(ResGridlet j, ResGridletList l)
  {
    if(this.remove(j))
    {
      l.add(j);
      return true;
    }
    return false;
  }
}

/**
 * A list of Grid Resource Users.
 */

class ResourceUserList extends LinkedList
{
  public ResourceUserList()
  {
    super();
  }

  public boolean add(int user_id)
  {
    Integer user = new Integer(user_id);
    if(this.contains(user))
      return false;
    this.add(user);
    return true;
  }

  public boolean myRemove(int user_id) {
    return(this.remove(new Integer(user_id)));
  }
}

/** It is a Resource Entity, which simulates time and space share machines.
 * The Resource can be a single CPU machines, SMP, or cluster.
 */
public class GridResource extends GridSim
{
  // |Machines| > 1 is cluster managed by queueing system
  private Random r;
  private ResourceCharacteristics resource;
  private ResourceCalendar rCalendar;
  private ResourceUserList users;
  private ResGridletList gridlets_arrived; // ready for execution
  private ResGridletList gridlets_in_execution;
  private ResGridletList gridlets_finished; // for accounting
  private double LastUpdateTime;
  private double LocalLoad;
  private int schedule_id;
  private Accumulator accTotalLoad;

  public GridResource(String name, double baud_rate, long seed,
                   ResourceCharacteristics resource, double peakLoad, double offpeakLoad,
                   double relativeHolidayLoad, LinkedList weekends,
                   LinkedList holydays)
  {
    super(name, baud_rate);
    this.r = new Random(seed);
    this.resource = resource;
    resource.SetID(get_id());
    this.rCalendar = new ResourceCalendar(resource.GetTimeZone(), peakLoad, offpeakLoad,
                   relativeHolidayLoad, weekends, holydays, seed);
    this.users = new ResourceUserList();
    this.gridlets_arrived = new ResGridletList();
    this.gridlets_in_execution = new ResGridletList();
    this.gridlets_finished = new ResGridletList();
    this.LastUpdateTime = Sim_system.clock();
    this.LocalLoad = rCalendar.CurrentLoad();
    this.schedule_id = 0;
    accTotalLoad = new Accumulator();
    accTotalLoad.add(TotalLoad());
  }

  public void body()
  {
    // RESOURCE REGISTRATION:
    Send(output, GridSimTags.SCHEDULE_NOW, GridSimTags.REGISTER_RESOURCE,
                  new IO_data(new Integer(get_id()), 4, GridSim.GridInformationServiceEntityId()));

    // Place Hourly Interrupt Request to itself: This is handle change in
    // load/access cost every hour (assumption)
    Calendar NextHour = rCalendar.CalendarNow();
    NextHour.add(Calendar.HOUR, 1);
    NextHour.set(Calendar.MINUTE, 0);
    NextHour.set(Calendar.SECOND, 0);
    Send(get_id(), rCalendar.SimulationTime(NextHour)-Sim_system.clock(), GridSimTags.INSIGNIFICANT);

    // Process Events until END_OF_SIMULATION is received from the GridSimShutdown Entity
    Sim_event ev = new Sim_event();
    for( sim_get_next(ev);  ev.get_tag() != GridSimTags.END_OF_SIMULATION; sim_get_next(ev) )
    {
      if(ev.get_src()==get_id())       // Internal Event if event source is the current enity
      {
        if(ev.get_data() != null)
        {
          if(ev.get_tag()==schedule_id) // A Gridlet is finished
            GridletFinishes((ResGridlet) ev.get_data());
          else
          {
            // the event is ignored as situation changed because of new gridlet arrival
            // as new completion time has already been scheduled
          }
        }
        else  // Hourly interrupt
          HourlyInterrupt();
      }
      else   // External Events
      {
        int src_id = -1;
        switch(ev.get_tag())
        {
          case GridSimTags.RESOURCE_CHARACTERISTICS: // Resource characteristics inquiry
            src_id = ((Integer) ev.get_data()).intValue();
            Send(output, GridSimTags.SCHEDULE_NOW, ev.get_tag(),
              new IO_data(resource, resource.ByteSize(), src_id));
            break;

          case GridSimTags.RESOURCE_DYNAMICS: // Resource dynamic info inquiry:
            src_id = ((Integer) ev.get_data()).intValue();
            switch (resource.GetAllocationPolicy())
            {
              case ResourceCharacteristics.TIME_SHARED:
                Send(output, GridSimTags.SCHEDULE_NOW, ev.get_tag(),
                  new IO_data(accTotalLoad, Accumulator.ByteSize(), src_id));
                break;
              case ResourceCharacteristics.SPACE_SHARED:
              case ResourceCharacteristics.ADVANCE_RESERVATION:
//                Send(output, GridSimTags.SCHEDULE_NOW, ev.get_tag(),
//									new IO_data(GetNumberOfFreePE(), 4, src_id));
                break;
            }
            break;
          case GridSimTags.GRIDLET_SUBMIT: // New gridlet arrives
            switch (resource.GetAllocationPolicy())
            {
              case ResourceCharacteristics.TIME_SHARED:
                GridletArrives((Gridlet) ev.get_data());
                break;
              case ResourceCharacteristics.SPACE_SHARED:
              /*
                1. Identify Unallocated PEs
                2. If some free PEs available, map gridlet to aavialble PE
                   a. Total Execution Time = [GridletMI/(PE's MPIS Rating)]+IO Time
                   b. Set an Event to be se sent after Total_Execution_Time
                      indicating completion of Gridlet
                  Else,
                     Place Gridlet into "Job Queue".
              */
                break;
            }
            break;
        }
      }
    }
    TerminateInputOutputEntities();
  }

  private double TotalLoad()
  {
    int NoOfGridletsPerPE = (int) Math.ceil((gridlets_in_execution.size()+1.0)/
                                            resource.GetNoOfPEs());
    return 1.0 - (1 - LocalLoad)/NoOfGridletsPerPE;
  }

  // Gridlets MI share in Time Shared Mode
  class MIShares
  {
    public double max; // max amount of MI share Gridlets can get
    public double min; // max amount of MI share Gridlets can get when few PEs need to run more Gridlets than other PEs
    public int maxCount; // Total number of Gridlets that get Max share.
  }

  /**
   * GetMIShare: Identifies MI share (max and min) each Gridlets gets for a given TimeSpan
   */
  private MIShares GetMIShare(double TimeSpan)
  {
    MIShares obj = new MIShares();
    double TotalMIperPE = resource.GetMIPSRatingOfOnePE()*TimeSpan*(1-LocalLoad); // 1 - LocalLoad = available MI share percentage
    int NoOfGridletsInExec = gridlets_in_execution.size();
    int NoOfPEs = resource.GetNoOfPEs();
    int glDIVpe = NoOfGridletsInExec / NoOfPEs;
    int glMODpe = NoOfGridletsInExec % NoOfPEs;

    if(glDIVpe>0)  // If NoOfGridletsInExec > NoOfPEs then, divide MIShare among all Gridlets
      obj.max = TotalMIperPE/(glDIVpe);
    else
      obj.max = TotalMIperPE;
    obj.min = TotalMIperPE/(glDIVpe+1);
    obj.maxCount = (NoOfPEs - glMODpe)*glDIVpe;

    return obj;
  }

  private void UpdateFinishSoFarTime()
  {
    double TimeSpan = Sim_system.clock() - LastUpdateTime;
    MIShares shares = GetMIShare(TimeSpan);

    LastUpdateTime = Sim_system.clock();
    LocalLoad = rCalendar.CurrentLoad();
    accTotalLoad.add(TotalLoad());

    for(int i = 0; i <gridlets_in_execution.size(); i++) {

      if(i < shares.maxCount)
        ((ResGridlet) gridlets_in_execution.get(i)).UpdateGridletFinishedSoFar(shares.max);
      else
        ((ResGridlet) gridlets_in_execution.get(i)).UpdateGridletFinishedSoFar(shares.min);
    }
  }

  private void UpdateFinishTime()
  {
    double TimeSpan = 1; // One second
    MIShares shares = GetMIShare(TimeSpan);

    for(int i = 0; i <gridlets_in_execution.size(); i++) {
      if(i < shares.maxCount)
        ((ResGridlet) gridlets_in_execution.get(i)).SetFinishTime(shares.max, rCalendar);
      else
        ((ResGridlet) gridlets_in_execution.get(i)).SetFinishTime(shares.min, rCalendar);
    }
  }

  /**
   * Identify time by which one of the Gridlet finishes and set that time
   * as a time for raising the next event.
   */
  private void ScheduleNextCompletionEvent()
  {
    ResGridlet MinGl = null;

    // FOR TIME SHARED: Identify a Gridlet which minimum remaining completion time
    if(gridlets_in_execution.size()>0)
    {
      for(int i = 0; i < gridlets_in_execution.size(); i++)
      {
        ResGridlet gl = (ResGridlet) gridlets_in_execution.get(i);
        if(gl.GetFinishTime() < 0)
          continue;
        if(MinGl==null)
          MinGl = gl;   //assume first small Gridlet
        else if(gl.GetFinishTime() < MinGl.GetFinishTime())
          MinGl = gl;   // smallest gridlet identified so far
      }
    }

    // FOR SPACE SHARED: Identify a Gridlet to be schedule for execution next
    /*
       CODE:
         1. Get a Gridlet
         2. Assign to PE
         3. Indicate that PE is busy
    */

    // Schedule Gridlet completion event
    if(MinGl!=null)
    {
      schedule_id++;
      Send(get_id(), MinGl.GetFinishTime()-Sim_system.clock(), schedule_id, MinGl);
    }
  }

  private void GridletArrives(Gridlet gl)
  {
    ResGridlet rgl;
    UpdateFinishSoFarTime();
    switch (resource.GetAllocationPolicy())
    {
      case ResourceCharacteristics.TIME_SHARED:
        gl.SetStatus(Gridlet.INEXEC);
        gl.SetExecStartTime(Sim_system.clock());
        rgl = new ResGridlet(gl);
        gridlets_in_execution.add(rgl);
        UpdateFinishTime();
        ScheduleNextCompletionEvent();
        break;
      case ResourceCharacteristics.SPACE_SHARED:
          gl.SetStatus(Gridlet.QUEUED);
          gl.SetExecStartTime(Sim_system.clock());
          // More LOGIC to follow
        break;
      case ResourceCharacteristics.ADVANCE_RESERVATION:
        break;
    }
  }

  private void GridletFinishes(ResGridlet rgl)
  {
    UpdateFinishSoFarTime();

    rgl.GetGridlet().SetStatus(Gridlet.SUCCESS);
    double WallClockTime = rgl.GetFinishTime()-rgl.GetArrivalTime();
    // calculate cpu time
    double CPUTime = 0;
    switch (resource.GetAllocationPolicy())
    {
      case ResourceCharacteristics.TIME_SHARED:
        // Only CPU consumed is measured since during I/O time, CPU might have been used
        // by other Gridlets
        CPUTime = rgl.GetGridlet().GetGridletLength()/resource.GetMIPSRatingOfOnePE();

        break;
      case ResourceCharacteristics.SPACE_SHARED:
      case ResourceCharacteristics.ADVANCE_RESERVATION:
        // CPU time == [Exec_EndTime - Exec_StartTime] since resource id dedicated.
        break;
    }

    rgl.GetGridlet().SetExecParam(WallClockTime, CPUTime);
    Send(output, GridSimTags.SCHEDULE_NOW, GridSimTags.GRIDLET_RETURN,
                 new IO_data(rgl.GetGridlet(), rgl.GetGridlet().GetGridletOutputSize(),
                              rgl.GetGridlet().GetUserID()));

    gridlets_in_execution.move(rgl, gridlets_finished);
    // System.out.println(get_name()+": I finished: "+gridlets_finished.size());

    UpdateFinishTime();
    ScheduleNextCompletionEvent();
  }

  /** It places request to raise hourly event to itself.
   */
  private void HourlyInterrupt()
  {
    UpdateFinishSoFarTime();
    UpdateFinishTime();
    ScheduleNextCompletionEvent();
    Send(get_id(), 60.0*60.0, GridSimTags.INSIGNIFICANT); // next hour
  }
}
