/**
 * Title:        GridSim Toolkit
 * Description:  GridSim (Grid Simulation) Toolkit for Modeling and Simulation of
 *               Parallel and Distributed Systems such as Clusters and Grids
 * @version      1.0, 12/09/2001 (September 12, 2001)
 * @version      1.0 28/12/2001 (updated by Raj)
 *
 * @author       R. Buyya and M. Murshed (by surname)
 *               Monash University, Melbourne, Australia
 * Licence:      GPL - http://www.gnu.org/copyleft/gpl.html
 */

package gridsim;
import java.io.*;
import java.util.*;
import eduni.simjava.*;

/** The top most GridSim class that all GridSim entities need to extend
 *  to become concurrent entities. They automatically inherit all
 *  properties of entity which extends thread class. All entities
 *  extending GridSim should defined a <tt>body()</tt> method inwhich
 *  entitity actions (e.g., raising, processing, and receiveing events to communicate
 *  with other entities) are performed.
 *  When the execution of an entity is initiated by <tt>Start</tt> method,
 *  the method <tt>body()</tt> is automatically invoked by entity thread.
 *  Entities should exit </tt>body()</tt> method when there are no more
 *  interested in handling events.
 */

public class GridSim extends Sim_entity
{
  // Entitiy with Network Communication
  protected Sim_port input;    // reading data received via input port
  protected Sim_port output;   // sending data via output port to external entities
  private double baud_rate;    // communication speed
  private boolean networkedFlag;  // true, if networked entity, other false
  private boolean TerminateInputOutputEntitiesFlag = false; // false means NOT invoked

  /** Entity WITHOUT NETWORK communication channels: "input" and "output" Sim_port.
   */
  public GridSim(String name)
  {
    super(name);
    networkedFlag = false;
    input = output = null;
  }

  /** Entitity with NETWORK communication channels: "input" and "output" Sim_port
   */
  public GridSim(String name, double baud_rate)
  {
    super(name);
    networkedFlag = true;

    this.baud_rate = baud_rate;
    input = new Sim_port("input");
    add_port(input);
    output = new Sim_port("output");
    add_port(output);

    // Every GridSim_entity_with_network has its own input/output channels.
    new Input("Input_"+name, baud_rate);
    // Connect this entity "input" port to its input's buffer's "in_port")
    Sim_system.link_ports(name, "input", "Input_"+name, "input_buffer");
    new Output("Output_"+name, baud_rate);
    Sim_system.link_ports(name, "output", "Output_"+name, "output_buffer");
  }

  /** It termindates Entities managing NETWORK communication channels
   * Can be invoked explicity to shutdown NETWORK communication channels.
   * It is advisable for for all entities extending GridSim class, explicityly
   * invote this method to terminate <tt>Input</tt> and <tt>Output</tt> entities created
   * by the constructor: GridSim(String name, double baud_rate)
   */
  protected void TerminateInputOutputEntities()
  {
    // If it is Networked entity and Not yet terminated, then terminate.
    if( IsNetworked() && !TerminateInputOutputEntitiesFlag)
    {
      Sim_event ev = new Sim_event();
      // Send END_OF_SIMULATION to Input entity
      Send(input, GridSimTags.SCHEDULE_NOW, GridSimTags.END_OF_SIMULATION);
      // Send END_OF_SIMULATION to Output entity
      Send(output, GridSimTags.SCHEDULE_NOW, GridSimTags.END_OF_SIMULATION);
      TerminateInputOutputEntitiesFlag = true;
//			System.out.println(get_name()+": IO terminated!");
    }
  }

  /** Invoked by the garbage collector to clearn up any data
   */
  protected void finalize()
  {
    TerminateInputOutputEntities();
  }

  /**
   * Check type of entity
   *  @return true if entity has NETWORK communication channel
   *  @return false, otherwise.
   */
  public boolean IsNetworked()
  {
    return networkedFlag;
  }

  /** Send Event/Message to Entity by 'delay' simulation time from now with
   * user-defined tag representing the type of event.
   * @param EntityName The name of the destination entity
   * @param delay  How long from the current simulation time the event/message should be delivered
   * @param GridSimTag  An user-defined number representing the type of event/message
   */
  protected void Send(String EntityName, double delay, int GridSimTag )
  {
    sim_schedule(GridSim.GetEntityId(EntityName), delay, GridSimTag);
  }

  // Send Event/Message with data
  protected void Send(String EntityName, double delay, int GridSimTag, Object data)
  {
    sim_schedule(GridSim.GetEntityId(EntityName), delay, GridSimTag, data);
  }

  protected void Send(int EntityID, double delay, int GridSimTag )
  {
    sim_schedule(EntityID, delay, GridSimTag);
  }

  protected void Send(int EntityID, double delay, int GridSimTag, Object data )
  {
    sim_schedule(EntityID, delay, GridSimTag, data);
  }

  protected void Send(Sim_port dest_port, double delay, int GridSimTag )
  {
    sim_schedule(dest_port, delay, GridSimTag);
  }

  protected void Send(Sim_port dest_port, double delay, int GridSimTag, Object data )
  {
    sim_schedule(dest_port, delay, GridSimTag, data);
  }

  protected Object ReceiveEventObject()
  {
    Sim_event ev = new Sim_event();
    sim_get_next(ev);
    return ev.get_data();
  }

  protected Object ReceiveEventObject(Sim_port source_port)
  {
    Sim_event ev = new Sim_event();
    sim_get_next( new Sim_from_port(source_port), ev);
    return ev.get_data();
  }

  //////////// SIMULATION CONTROL METHODS START ////////////

  public static Date SimulationStartDate; // w.r.t. GMT 0

  public static Date GetSimulationStartDate()
  {
    return SimulationStartDate;
  }

  public static GridSimRandom rand;

  /** Initialise GridSim Parameters.
   *  This should be called before creating any entities.
   *  @param no_of_users The number of User Entities created. This parameters indicates that
   *  GridSimShutdown first waits for User Entities's END_OF_SIMULATION signal before issuing
   *  terminate singal to other entities.
   *  @param cal The calender object for storing peaktime, holiday, etc. time
   *  @param traceFlag true if GridSim trace need to be written.
   *  @param exclude_from_file
   *  @param exclude_from_processing
   *  @param ReportWriterName The name of the entities which is signaled to write
   *  write any specific report just before termination of the simulation.
   */
  public static void Init(int no_of_users, Calendar cal, boolean traceFlag,
          String[] exclude_from_file, String[] exclude_from_processing, String ReportWriterName)
  {
    Sim_system.initialise();
    Sim_system.set_auto_trace(traceFlag);
    Sim_system.set_trc_level(1);

    SimulationStartDate = cal.getTime();
    rand = new GridSimRandom();

    new GridStatistics("GridStatistics", "stat.txt", true, exclude_from_file, exclude_from_processing);
    new GridSimShutdown("GridSimShutdown", no_of_users, ReportWriterName);
    new GridInformationService("GridInformationService", GridSimTags.DEFAULF_BAUD_RATE);
  }

  /** Start the execution of GridSim simulation.
   *  It waits for complete execution of all entitities, i.e., until
   *  all entities thread reach non-RUNNABLE state by exiting from the <tt>body()</tt> method.
   *  Then, kills threads of all entities.
   *  <b>Note</b>: This method should be called after all the entities have been setup and
   *  added, and their ports linked.
   */
  public static void Start()
  {
      // author names arranged by surname since both developers put good amount of energy for coding!
      System.out.println("GridSim V 1.0 -- R. Buyya and M. Murshed!");
      Sim_system.run();
  }

  /**
   * Get Current Simulation Time.
   * @return The current simulation time from the simulation clock
   */
  public static double Clock()
  {
    return Sim_system.clock();
  }

  /** Causes the entity to hold for <tt>duration</tt> units of simulation time.
   * @param duration The amount of time to hold.
   */
  public void GridSimHold( double duration )
  {
    sim_hold( duration );
  }

  /** Stop Grid Simulation.
   *  This should be called only any of the user defined entities <tt>Explicitly</tt> want
   *  to terminate simulation during execution.
   */
  public static void Stop()
  {
    Sim_system.run_stop();
  }
  ////////////  SIMULATION CONTROL METHODS END ////////////


 ////////////  ENTITY ID/NAME CONVERSION METHODS START ////////////

  /** Returns name of the entity given its Entity ID
   */
  public static String GetEntityName( int EntityID )
  {
      return (String) Sim_system.get_entity(EntityID).get_name();
  }

  /** It returns ID (int value) of the Entity given its name
   */
  public static int GetEntityId(String EntityName )
  {
     return (int) Sim_system.get_entity_id(EntityName);
  }

  /** It returns ID of the Entity named "GridStatistics"
   */
  public static int GridStatisticsEntityId()
  {
     return (int) Sim_system.get_entity_id("GridStatistics");
  }

  /** It returns ID of the Entity named "GridInformationService"
   */
  public static int GridInformationServiceEntityId()
  {
     return (int) Sim_system.get_entity_id("GridInformationService");
  }

  /** It returns ID of Entity named "GridSimShutdown"
   */
  public static int GridSimShutdownEntityId()
  {
     return (int) Sim_system.get_entity_id("GridSimShutdown");
  }

  /** This should be invoved by all User Entities.
   * GridSimShutdownEntity waits for all Users termination before shuting down other entities.
   */
  protected void ShutdownUserEntity()
  {
    Send(GridSim.GridSimShutdownEntityId(), GridSimTags.SCHEDULE_NOW, GridSimTags.END_OF_SIMULATION);
  }

  /** This method should be invoked by User Defined ReportWriter
   * if report writter is NOY defined (null) during Init, then GridSimShutDown will invoke it
   */
  protected void ShutdownGridStatisticsEntity()
  {
    Send(GridSim.GridStatisticsEntityId(), GridSimTags.SCHEDULE_NOW, GridSimTags.END_OF_SIMULATION);
  }

  /**
   * It send sends a request to GIS entity to get the list of Grid resources.
   * @returns A LinkedList containing GridResourceId as Integer Object.
   */
  public LinkedList GetGridResourceList()
  {
      LinkedList ResIDList;
      // Requesting the GridInformationService to send a list of resources
      // To "IO_data", my entity ID, size of data transfer, and destination entity ID is passed.
      Send(output, GridSimTags.SCHEDULE_NOW, GridSimTags.RESOURCE_LIST,
          new IO_data(new Integer(get_id()), 4, GridSim.GridInformationServiceEntityId()));
      ResIDList = (LinkedList) ReceiveEventObject();
      return ResIDList;
  }
  ////////////  ENTITY ID/NAME CONVERSION METHODS END ////////////

  public void RecordStatistics(String category, double data)
  {
    Send(GridStatisticsEntityId(), GridSimTags.SCHEDULE_NOW,
      GridSimTags.RECORD_STATISTICS, new Stat(Sim_system.clock(), category, get_name(), data));
  }
  public void RecordStatistics(String category, int data)
  {
    Send(GridStatisticsEntityId(), GridSimTags.SCHEDULE_NOW,
      GridSimTags.RECORD_STATISTICS, new Stat(Sim_system.clock(), category, get_name(), data));
  }
  public void RecordStatistics(String category, String data)
  {
    Send(GridStatisticsEntityId(), GridSimTags.SCHEDULE_NOW,
      GridSimTags.RECORD_STATISTICS, new Stat(Sim_system.clock(), category, get_name(), data));
  }
  public void RecordStatistics(String category, boolean data)
  {
    Send(GridStatisticsEntityId(), GridSimTags.SCHEDULE_NOW,
      GridSimTags.RECORD_STATISTICS, new Stat(Sim_system.clock(), category, get_name(), data));
  }
}
