/**
 * Title:        GridSim Toolkit
 * Version:      GridSim  1.0, 12/09/2001
 * Description:  GridSim (Grid Simulation) Toolkit for Modeling and Simulation of
 *               Parallel and Distributed Systems such as Clusters and Grids
 * @author       R. Buyya and M. Murshed (by surname)
 *               Monash University, Melbourne, Australia
 * Licence:      GPL - http://www.gnu.org/copyleft/gpl.html
 */

package gridsim;
import java.util.*;
import java.io.*;
import eduni.simjava.*;

/**
 * A resource characteristics information.
 */
public class ResourceCharacteristics
{
  private int id; // resource id--created internally
  private String architecture;
  private String OS; // Example: "Linux" same as result of unix command "uname"
  private MachineList machineList;
  private int allocation_policy; // 0 = time shared; 1 = space shared; 2 = 1+advance reservation
  private double TimeZone; // difference from GMT
  private double CostPerSec;  // Price/CPU-unit if unit = sec., then G$/CPU-sec.

  // Resource Types
  public static final int TIME_SHARED = 0;
  public static final int SPACE_SHARED = 1;
  public static final int ADVANCE_RESERVATION = 2;

  public ResourceCharacteristics(String architecture, String OS, MachineList machineList,
                      int allocation_policy, double TimeZone, double CostPerSec) {
    this.architecture = architecture;
    this.OS = OS;
    this.machineList = machineList;
    this.allocation_policy = allocation_policy;
    this.TimeZone = TimeZone;
    this.CostPerSec = CostPerSec;
  }

  /** Set Resource ID
   */
  public void SetID(int id)
  {
    this.id = id;
  }

  /** Get Resource ID
   */
  public int GetID()
  {
    return id;
  }

  public String GetArch()
  {
    return architecture;
  }

  public String GetOS()
  {
    return OS;
  }

  public MachineList GetMachines()
  {
    return machineList;
  }

  public int GetAllocationPolicy()
  {
    return allocation_policy;
  }

  public double GetTimeZone()
  {
    return TimeZone;
  }

  /** Get MIPS Rating of a PE.
   *  It is assumed all PE's rating is same in a given Machine.
   */
  public int GetMIPSRatingOfOnePE()
  {
    return machineList.GetMachine(0).GetPEs().GetMIPSRating(0);
  }
  /** Get MIPS Rating of a PE.
   *  It is essential to use this method when a resource is made up of heterogenous
   *  PEs/machines.
   */
  public int GetMIPSRatingOfOnePE(int machine_ID, int PE_ID)
  {
    return machineList.GetMachine(machine_ID).GetPEs().GetMIPSRating(PE_ID);
  }

  public int GetMIPSRating()
  {
    int rating = 0;
    switch(allocation_policy)
    {
      case ResourceCharacteristics.TIME_SHARED:
        // Assuming all PEs in a Machine have same rating.
        rating = GetMIPSRatingOfOnePE()*machineList.GetNoOfPEs();
        break;
      case ResourceCharacteristics.SPACE_SHARED:
      case ResourceCharacteristics.ADVANCE_RESERVATION:
        // Assuming all PEs in a Machine have same rating.
        // But different machines in a Cluster can have different rating.
        for(int i=0; i<machineList.size(); i++)
          rating += ((Machine) machineList.get(i)).GetMIPSRating();
        break;
    }
    return rating;
  }

  public double CPU_time(double GridletLength, double load)
  {
    double cpuTime = 0.0;
    switch(allocation_policy) {
      case ResourceCharacteristics.TIME_SHARED:
        cpuTime = GridletLength/(GetMIPSRatingOfOnePE()*(1.0-load));
        break;
      case ResourceCharacteristics.SPACE_SHARED:
      case ResourceCharacteristics.ADVANCE_RESERVATION:
//        cpuTime = GridletLength/(GetMIPSRatingOfOnePE(machine_id, PE_id)*(1.0-load));
        break;
    }
    return cpuTime;
  }

  public int GetNoOfPEs() {
    return machineList.GetNoOfPEs();
  }

  public void SetCostPerSec(double CostPerSec) {
    this.CostPerSec = CostPerSec;
  }

  public double GetCostPerSec() {
    return CostPerSec;
  }

  public double GetCostPerMI() {
    return CostPerSec/GetMIPSRatingOfOnePE();
  }

  public int ByteSize() {
    return architecture.length() + OS.length() + machineList.ByteSize() +
      4 + 2*8;  // sizeof(allocation_policy) + sizeof(TimeZone)
  }
}
